@@ -18,6 +18,7 @@
 #include "config.h"
 
 static int cache_timeout;
+static char *temp_dir;
 
 typedef struct dir_cache
 {
@@ -86,7 +87,7 @@ static int caching_list_directory(const char *path, dir_entry **list)
   return 1;
 }
 
-static void update_dir_cache(const char *path, off_t size, int isdir)
+static void update_dir_cache(const char *path, off_t size, int isdir, int islink)
 {
   pthread_mutex_lock(&dmut);
   dir_cache *cw;
@@ -109,9 +110,23 @@ static void update_dir_cache(const char *path, off_t size, int isdir)
       de = (dir_entry *)malloc(sizeof(dir_entry));
       de->size = size;
       de->isdir = isdir;
+      de->islink = islink;
       de->name = strdup(&path[strlen(cw->path)+1]);
       de->full_name = strdup(path);
-      de->content_type = strdup(isdir ? "application/directory" : "application/octet-stream");
+
+      if (isdir)
+      {
+        de->content_type = strdup("application/link");
+      }
+      if(islink)
+      {
+        de->content_type = strdup("application/directory");
+      }
+      else
+      {
+
+        de->content_type = strdup("application/octet-stream");
+      }
       de->last_modified = time(NULL);
       de->next = cw->entries;
       cw->entries = de;
@@ -204,6 +219,15 @@ static int cfs_getattr(const char *path, struct stat *stbuf)
     stbuf->st_mode = S_IFDIR | 0755;
     stbuf->st_nlink = 2;
   }
+  else if (de->islink)
+  {
+    stbuf->st_size = 1;
+    stbuf->st_mode = S_IFLNK | 0755;
+    stbuf->st_nlink = 1;
+    stbuf->st_size = de->size;
+    /* calc. blocks as if 4K blocksize filesystem; stat uses units of 512B */
+    stbuf->st_blocks = ((4095 + de->size) / 4096) * 8;
+  }
   else
   {
     stbuf->st_size = de->size;
@@ -244,7 +268,7 @@ static int cfs_mkdir(const char *path, mode_t mode)
 {
   if (cloudfs_create_directory(path))
   {
-    update_dir_cache(path, 0, 1);
+    update_dir_cache(path, 0, 1, 0);
     return 0;
   }
   return -ENOENT;
@@ -258,24 +282,90 @@ static int cfs_create(const char *path, mode_t mode, struct fuse_file_info *info
   fclose(temp_file);
   of->flags = info->flags;
   info->fh = (uintptr_t)of;
-  update_dir_cache(path, 0, 0);
+  update_dir_cache(path, 0, 0, 0);
   info->direct_io = 1;
   return 0;
 }
 
 static int cfs_open(const char *path, struct fuse_file_info *info)
 {
-  FILE *temp_file = tmpfile();
+  //TODO: need to clean this up and reduce the duplicated code
+  if (!*temp_dir) {
+
+    FILE *temp_file = tmpfile();
+    dir_entry *de = path_info(path);
+    if (!(info->flags & O_WRONLY))
+    {
+      if (!cloudfs_object_write_fp(path, temp_file))
+      {
+        fclose(temp_file);
+        return -ENOENT;
+      }
+      update_dir_cache(path, (de ? de->size : 0), 0, 0);
+    }
+    openfile *of = (openfile *)malloc(sizeof(openfile));
+    of->fd = dup(fileno(temp_file));
+    fclose(temp_file);
+    of->flags = info->flags;
+    info->fh = (uintptr_t)of;
+    info->direct_io = 1;
+    return 0;
+  }
+
+
+  FILE *temp_file;
+
+  char tmp_path[PATH_MAX];
+  strncpy(tmp_path, path, PATH_MAX);
+
+  char *pch;
+  while((pch = strchr(tmp_path, '/'))) {
+    *pch = '.';
+  }
+
+  char file_path[PATH_MAX];
+  snprintf(file_path, PATH_MAX, "%s/.cloudfuse%ld-%s", temp_dir,
+      (long)getpid(), tmp_path);
+
   dir_entry *de = path_info(path);
-  if (!(info->flags & O_WRONLY))
-  {
+
+  if(access(file_path, F_OK) != -1) {
+    temp_file = fopen(file_path, "r");
+    update_dir_cache(path, (de ? de->size : 0), 0, 0);
+    // file exists
+  } else if (!(info->flags & O_WRONLY)) {
+    // we need to lock on the filename another process could open the file
+    // while we are writing to it and then only read part of the file
+
+    // duplicate the directory caching datastructure to make the code easier
+    // to understand.
+
+    // each file in the cache needs:
+    //  filename, is_writing, last_closed, is_removing
+    // the first time a file is opened a new entry is created in the cache
+    // setting the filename and is_writing to true.  This check needs to be
+    // wrapped with a lock.
+    //
+    // each time a file is closed we set the last_closed for the file to now
+    // and we check the cache for files whose last
+    // closed is greater than cache_timeout, then start a new thread rming
+    // that file.
+
+    // TODO: just to prevent this craziness for now
+    if (*temp_dir)
+        temp_file = fopen(file_path, "w+b");
+    else
+        temp_file = tmpfile();
+
+
     if (!cloudfs_object_write_fp(path, temp_file))
     {
       fclose(temp_file);
       return -ENOENT;
     }
-    update_dir_cache(path, (de ? de->size : 0), 0);
+    update_dir_cache(path, (de ? de->size : 0), 0, 0);
   }
+
   openfile *of = (openfile *)malloc(sizeof(openfile));
   of->fd = dup(fileno(temp_file));
   fclose(temp_file);
@@ -295,7 +385,7 @@ static int cfs_flush(const char *path, struct fuse_file_info *info)
   openfile *of = (openfile *)(uintptr_t)info->fh;
   if (of)
   {
-    update_dir_cache(path, cloudfs_file_size(of->fd), 0);
+    update_dir_cache(path, cloudfs_file_size(of->fd), 0, 0);
     if (of->flags & O_RDWR || of->flags & O_WRONLY)
     {
       FILE *fp = fdopen(dup(of->fd), "r");
@@ -319,7 +409,10 @@ static int cfs_release(const char *path, struct fuse_file_info *info)
 
 static int cfs_rmdir(const char *path)
 {
-  if (cloudfs_delete_object(path))
+  int success = cloudfs_delete_object(path);
+  if (success == -1)
+    return -ENOTEMPTY;
+  if (success)
   {
     dir_decache(path);
     return 0;
@@ -333,19 +426,22 @@ static int cfs_ftruncate(const char *path, off_t size, struct fuse_file_info *in
   if (ftruncate(of->fd, size))
     return -errno;
   lseek(of->fd, 0, SEEK_SET);
-  update_dir_cache(path, size, 0);
+  update_dir_cache(path, size, 0, 0);
   return 0;
 }
 
 static int cfs_write(const char *path, const char *buf, size_t length, off_t offset, struct fuse_file_info *info)
 {
-  update_dir_cache(path, offset + length, 0);
+  update_dir_cache(path, offset + length, 0, 0);
   return pwrite(((openfile *)(uintptr_t)info->fh)->fd, buf, length, offset);
 }
 
 static int cfs_unlink(const char *path)
 {
-  if (cloudfs_delete_object(path))
+  int success = cloudfs_delete_object(path);
+  if (success == -1)
+    return -EACCES;
+  if (success)
   {
     dir_decache(path);
     return 0;
@@ -393,12 +489,41 @@ static int cfs_rename(const char *src, const char *dst)
   if (cloudfs_copy_object(src, dst))
   {
     /* FIXME this isn't quite right as doesn't preserve last modified */
-    update_dir_cache(dst, src_de->size, 0);
+    update_dir_cache(dst, src_de->size, 0, 0);
     return cfs_unlink(src);
   }
   return -EIO;
 }
 
+static int cfs_symlink(const char *src, const char *dst)
+{
+  if(cloudfs_create_symlink(src, dst))
+  {
+    update_dir_cache(dst, 1, 0, 1);
+    return 0;
+  }
+  return -EIO;
+}
+
+static int cfs_readlink(const char* path, char* buf, size_t size)
+{
+  FILE *temp_file = tmpfile();
+  int ret = 0;
+
+  if (!cloudfs_object_write_fp(path, temp_file))
+  {
+      ret = -ENOENT;
+  }
+
+  if (!pread(fileno(temp_file), buf, size, 0))
+  {
+      ret = -ENOENT;
+  }
+
+  fclose(temp_file);
+  return ret;
+}
+
 static void *cfs_init(struct fuse_conn_info *conn)
 {
   signal(SIGPIPE, SIG_IGN);
@@ -417,24 +542,31 @@ char *get_home_dir()
 }
 
 FuseOptions options = {
-    .username = "",
-    .password = "",
     .cache_timeout = "600",
     .verify_ssl = "true",
+    .segment_size = "1073741824",
+    .segment_above = "2147483648",
+    .storage_url = "",
+    .container = "",
+    //.temp_dir = "/tmp/",
+    .temp_dir = "",
     .client_id = "",
     .client_secret = "",
-    .redirect_uri = ""
+    .refresh_token = ""
 };
 
 int parse_option(void *data, const char *arg, int key, struct fuse_args *outargs)
 {
-  if (sscanf(arg, " username = %[^\r\n ]", options.username) ||
-      sscanf(arg, " password = %[^\r\n ]", options.password) ||
-      sscanf(arg, " cache_timeout = %[^\r\n ]", options.cache_timeout) ||
+  if (sscanf(arg, " cache_timeout = %[^\r\n ]", options.cache_timeout) ||
       sscanf(arg, " verify_ssl = %[^\r\n ]", options.verify_ssl) ||
+      sscanf(arg, " segment_above = %[^\r\n ]", options.segment_above) ||
+      sscanf(arg, " segment_size = %[^\r\n ]", options.segment_size) ||
+      sscanf(arg, " storage_url = %[^\r\n ]", options.storage_url) ||
+      sscanf(arg, " container = %[^\r\n ]", options.container) ||
+      sscanf(arg, " temp_dir = %[^\r\n ]", options.temp_dir) ||
       sscanf(arg, " client_id = %[^\r\n ]", options.client_id) ||
       sscanf(arg, " client_secret = %[^\r\n ]", options.client_secret) ||
-      sscanf(arg, " redirect_uri = %[^\r\n ]", options.redirect_uri))
+      sscanf(arg, " refresh_token = %[^\r\n ]", options.refresh_token))
     return 0;
   if (!strcmp(arg, "-f") || !strcmp(arg, "-d") || !strcmp(arg, "debug"))
     cloudfs_debug(1);
@@ -447,7 +579,7 @@ int main(int argc, char **argv)
   FILE *settings;
   struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
 
-  snprintf(settings_filename, sizeof(settings_filename), "%s/.cloudfuse", get_home_dir());
+  snprintf(settings_filename, sizeof(settings_filename), "%s/.hubicfuse", get_home_dir());
   if ((settings = fopen(settings_filename, "r")))
   {
     char line[OPTION_SIZE];
@@ -459,17 +591,29 @@ int main(int argc, char **argv)
   fuse_opt_parse(&args, &options, NULL, parse_option);
 
   cache_timeout = atoi(options.cache_timeout);
+  segment_size = atoll(options.segment_size);
+  segment_above = atoll(options.segment_above);
+  // this is ok since main is on the stack during the entire execution
+  override_storage_url = options.storage_url;
+  public_container = options.container;
+  temp_dir = options.temp_dir;
 
-  if (!*options.username || !*options.password)
+  if (!*options.client_id || !*options.client_secret || !*options.refresh_token)
   {
-    fprintf(stderr, "Unable to determine username and password.\n\n");
+    fprintf(stderr, "Unable to determine client_id, client_secret or refresh_token.\n\n");
     fprintf(stderr, "These can be set either as mount options or in "
                     "a file named %s\n\n", settings_filename);
-    fprintf(stderr, "  username=[Account username]\n");
-    fprintf(stderr, "  password=[Account password]\n");
+    fprintf(stderr, "  client_id=[App's id]\n");
+    fprintf(stderr, "  client_secret=[App's secret]\n");
+    fprintf(stderr, "  refresh_token=[Get it running hubic_token]\n");
     fprintf(stderr, "The following settings are optional:\n\n");
     fprintf(stderr, "  cache_timeout=[Seconds for directory caching, default 600]\n");
     fprintf(stderr, "  verify_ssl=[False to disable SSL cert verification]\n");
+    fprintf(stderr, "  segment_size=[Size to use when creating DLOs, default 1073741824]\n");
+    fprintf(stderr, "  segment_above=[File size at which to use segments, defult 2147483648]\n");
+    fprintf(stderr, "  storage_url=[Storage URL for other tenant to view container]\n");
+    fprintf(stderr, "  container=[Public container to view of tenant specified by storage_url]\n");
+    fprintf(stderr, "  temp_dir=[Directory to store temp files]\n");
 
     return 1;
   }
@@ -478,7 +622,8 @@ int main(int argc, char **argv)
 
   cloudfs_verify_ssl(!strcasecmp(options.verify_ssl, "true"));
 
-  cloudfs_set_credentials(options.username, options.password);
+  cloudfs_set_credentials(options.client_id, options.client_secret, options.refresh_token);
+
   if (!cloudfs_connect())
   {
     fprintf(stderr, "Failed to authenticate.\n");
@@ -510,6 +655,8 @@ int main(int argc, char **argv)
     .chmod = cfs_chmod,
     .chown = cfs_chown,
     .rename = cfs_rename,
+    .symlink = cfs_symlink,
+    .readlink = cfs_readlink,
     .init = cfs_init,
   };
 
